/*
* JBoss, Home of Professional Open Source.
* Copyright 2006, Red Hat Middleware LLC, and individual contributors
* as indicated by the @author tags. See the copyright.txt file in the
* distribution for a full listing of individual contributors. 
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/ 
package org.jboss.reflect.plugins.javassist.bytecode.benchmark;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.security.ProtectionDomain;

import javassist.CannotCompileException;
import javassist.Modifier;
import javassist.bytecode.AccessFlag;
import javassist.bytecode.Bytecode;
import javassist.bytecode.ClassFile;
import javassist.bytecode.ClassFileWriter;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.ConstPool;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.Opcode;
import javassist.bytecode.StackMapTable;
import javassist.util.proxy.FactoryHelper;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;


/**
 * 
 * @author <a href="kabir.khan@jboss.com">Kabir Khan</a>
 * @version $Revision: 1.1 $
 */
public class TestAsm implements Opcodes
{
   private static java.lang.reflect.Method defineClass1, defineClass2;

   static {
       try {
           Class<?> cl = Class.forName("java.lang.ClassLoader");
           defineClass1 = SecurityActions.getDeclaredMethod(
                       cl,
                       "defineClass",
                       new Class[] { String.class, byte[].class,
                                     int.class, int.class });

           defineClass2 = SecurityActions.getDeclaredMethod(
                       cl,
                       "defineClass",
                       new Class[] { String.class, byte[].class,
                             int.class, int.class, ProtectionDomain.class });
       }
       catch (Exception e) {
           throw new RuntimeException("cannot initialize");
       }
   }

   /** The number of methods handled for JavassistMethods impl (JavassistMethod just handles one) */ 
   private static final int BRANCHES = 1;
   
   public static void main(String[] args) throws Throwable
   {
      int max = 20000;
      
      TestAsm test = new TestAsm();

      Class<?> clazz = null;
      JavassistMethod m = null;
      JavassistMethods m2 = null;
      
      //Javassist and ASM versions of the JavassistMethod implementation
      clazz = test.generateAsmClassForJavassistMethod(0);
//      testInstantiateClassAndInvoke(clazz);
      
      clazz = test.generateJavassistClassForJavassistMethod(0);
//      testInstantiateClassAndInvoke(clazz);
      
      clazz = test.generateNewJvstClassForJvstMethod(0);
//      testInstantiateClassAndInvoke(clazz);


//      for(int k = 0; k < 5; k++){
      long asm = 0;
      long javassist = 0;
      long newJvst = 0;
      long start = 0;
      for (int i = 1 ; i < max ; i++)
      {
         start = System.currentTimeMillis();
         clazz = test.generateNewJvstClassForJvstMethod(i);
//         m = (JavassistMethod)clazz.newInstance();
         newJvst += System.currentTimeMillis() - start;

         //Javassist and ASM versions of the JavassistMethod implementation
         start = System.currentTimeMillis();
         clazz = test.generateJavassistClassForJavassistMethod(i);
//         m = (JavassistMethod)clazz.newInstance();
         javassist += System.currentTimeMillis() - start;

         start = System.currentTimeMillis();
         clazz = test.generateAsmClassForJavassistMethod(i);
//         m = (JavassistMethod)clazz.newInstance();
         asm += System.currentTimeMillis() - start;
      }
      
      System.out.println("Iterations\t" + max + "\t" + asm + "\t" + javassist + "\t" + newJvst + ": " + ((double)javassist) / asm + "\t" + newJvst + ": " + ((double)newJvst) / asm);
//      }
   }
   
   private static void testInstantiateClassAndInvoke(Class<?> clazz) throws Throwable
   {
      JavassistMethod m = (JavassistMethod)clazz.newInstance();
      if (m.invoke(new POJO(), new Object[] {"hello", 5}) != Integer.valueOf(5))
      throw new RuntimeException();
   }
   
   private ClassWriter generateAsmSkeleton(int i, String name, String superclass, String...interfaces)
   {
      ClassWriter cw = new ClassWriter(0);
      cw.visit(V1_5, ACC_PUBLIC, name, null, superclass, interfaces);
      // creates a MethodWriter for the (implicit) constructor
      MethodVisitor mw = cw.visitMethod(ACC_PUBLIC,
              "<init>",
              "()V",
              null,
              null);
      // pushes the 'this' variable
      mw.visitVarInsn(ALOAD, 0);
      // invokes the super class constructor
      mw.visitMethodInsn(INVOKESPECIAL, superclass, "<init>", "()V");
      mw.visitInsn(RETURN);
      // this code uses a maximum of one stack element and one local variable
      mw.visitMaxs(1, 1);
      mw.visitEnd();
      
      return cw;
   }
   
   private Class<?> generateAsmClassForJavassistMethod(int i) throws Exception
   {
      String realName = "pkg.ExampleAsmClass" + i;
      String name = "pkg/ExampleAsmClass" + i; 
      String superclass = "sun/reflect/MagicAccessorImpl";
      String[] interfaces = new String[] {"org/jboss/reflect/plugins/javassist/bytecode/benchmark/JavassistMethod"};
      
      ClassWriter cw = generateAsmSkeleton(i, name, superclass, interfaces);
      
      //Creates a MethodWriter for the invoke() method
      MethodVisitor mw = cw.visitMethod(ACC_PUBLIC, 
            "invoke", 
            "(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;", 
            null, 
            new String[] {"java/lang/Throwable"});

      //Load target
      mw.visitVarInsn(ALOAD, 1);
      mw.visitTypeInsn(CHECKCAST, "org/jboss/reflect/plugins/javassist/bytecode/benchmark/POJO");
      
      //First param
      mw.visitVarInsn(ALOAD, 2);
      mw.visitInsn(ICONST_0);
      mw.visitInsn(AALOAD);

      //Second param
      mw.visitVarInsn(ALOAD, 2);
      mw.visitInsn(ICONST_1);
      mw.visitInsn(AALOAD);
      mw.visitTypeInsn(CHECKCAST, "java/lang/Integer");
      mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Integer", "intValue", "()I");
      mw.visitMethodInsn(INVOKEVIRTUAL, "org/jboss/reflect/plugins/javassist/bytecode/benchmark/POJO", "accessInt", "(Ljava/lang/String;I)I");

      //Return
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Integer", "valueOf", "(I)Ljava/lang/Integer;");
      mw.visitInsn(ARETURN);
      
      
      mw.visitMaxs(1, 3);
      mw.visitEnd();
      
      byte[] bytes = cw.toByteArray();
      return null;
//      return toClass(realName, bytes, Thread.currentThread().getContextClassLoader(), null);
   }

    // begin new Javassist

    private void generateNewJvstSkeleton(ClassFileWriter.MethodWriter mw, int superclass, String superName)
        throws Exception
    {
        mw.begin(AccessFlag.PUBLIC, MethodInfo.nameInit, "()V", null, null);
        mw.add(Opcode.ALOAD_0);
        mw.addInvoke(Opcode.INVOKESPECIAL, superName, MethodInfo.nameInit, "()V");
        mw.add(Opcode.RETURN);
        mw.codeEnd(1, 1);
        mw.end(null, null);
    }

   private Class<?> generateNewJvstClassForJvstMethod(int i) throws Exception
   {
      String realName = "pkg.ExampleNewJavassistClass" + i;
      String name = "pkg/ExampleNewJavassistClass" + i;
      
      String superclass = "sun/reflect/MagicAccessorImpl";
      String[] interfaces
          = new String[] { "org/jboss/reflect/plugins/javassist/bytecode/benchmark/JavassistMethod" };

      ClassFileWriter cfw = new ClassFileWriter(ClassFile.JAVA_4, 0);
      ClassFileWriter.ConstPoolWriter cpw = cfw.getConstPool();

      int thisClass = cpw.addClassInfo(name);
      int superClass = cpw.addClassInfo(superclass);
      int[] interfaceTypes = cpw.addClassInfo(interfaces);

      ClassFileWriter.MethodWriter mw = cfw.getMethodWriter();

      generateNewJvstSkeleton(mw, superClass, superclass);

      mw.begin(AccessFlag.PUBLIC, "invoke", "(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;", new String[] { "java/lang/Throwable" }, null);

      //Target
      mw.add(Opcode.ALOAD_1);

      //First param
      mw.add(Opcode.ALOAD_2);
      mw.add(Opcode.ICONST_0);
      mw.add(Opcode.AALOAD);
      mw.add(Opcode.CHECKCAST);
      mw.add16(cpw.addClassInfo("java/lang/String"));

      mw.add(Opcode.ALOAD_2);
      mw.add(Opcode.ICONST_1);
      mw.add(Opcode.AALOAD);
      mw.add(Opcode.CHECKCAST);
      mw.add16(cpw.addClassInfo("java/lang/Integer"));
      mw.addInvoke(Opcode.INVOKEVIRTUAL, "java/lang/Integer", "intValue", "()I");

      mw.addInvoke(Opcode.INVOKEVIRTUAL,
                   "org/jboss/reflect/plugins/javassist/bytecode/benchmark/POJO",
                   "accessInt", "(Ljava/lang/String;I)I");

      //return
      mw.addInvoke(Opcode.INVOKESTATIC,
                   "java/lang/Integer", "valueOf", "(I)Ljava/lang/Integer;");
      mw.add(Opcode.ARETURN);
      mw.codeEnd(2, 3);

      // StackMapTable.Writer stack = new StackMapTable.Writer(32);
      // stack.sameFrame(1);

      mw.end(null, null);    // mw.end(stack, null);

      byte[] bytes = cfw.end(AccessFlag.PUBLIC, thisClass, superClass,
                             interfaceTypes, null);
      return null;
//      return toClass(realName, bytes, Thread.currentThread().getContextClassLoader(), null);
   }

    // end of new Javassist

   private ClassFile generateJavassistSkeleton(int i, String name, String superclass, String...interfaces) throws Exception
   {
      ClassFile cf = new ClassFile(false, name, superclass);
      cf.setAccessFlags(AccessFlag.PUBLIC);
      cf.setInterfaces(interfaces);
      ConstPool cp = cf.getConstPool();
      
      //Default ctor
      MethodInfo minfo = new MethodInfo(cp, "<init>", "()V");

      minfo.setAccessFlags(Modifier.PUBLIC);
      Bytecode code = new Bytecode(cp, 0, 0);
      int pc = code.currentPc();
      code.addAload(0);
      code.addInvokespecial(superclass, "<init>", "()V");
      code.addOpcode(Opcode.RETURN);
      code.setMaxLocals(1);
      CodeAttribute ca = code.toCodeAttribute();
      minfo.setCodeAttribute(ca);
      //StackMapTable.Writer writer = new StackMapTable.Writer(32);
      //writer.sameFrame(pc);
      //ca.setAttribute(writer.toStackMapTable(cp));
      cf.addMethod2(minfo);

      return cf;
   }

    static final String pojoName = "org/jboss/reflect/plugins/javassist/bytecode/benchmark/POJO";
   
   private Class<?> generateJavassistClassForJavassistMethod(int i) throws Exception
   {
      String name = "pkg.ExampleJavassistClass" + i; 
      String superclass = "sun/reflect/MagicAccessorImpl";
      String[] interfaces = new String[] {"org/jboss/reflect/plugins/javassist/bytecode/benchmark/JavassistMethod"};
      
      ClassFile cf = generateJavassistSkeleton(i, name, superclass, interfaces);
      
      //invoke Method
      ConstPool cp = cf.getConstPool();
      MethodInfo minfo = new MethodInfo(cp, "invoke", "(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;");

      minfo.setAccessFlags(Modifier.PUBLIC);
      Bytecode code = new Bytecode(cp, 0, 0);
      int pc = code.currentPc();

      //Target
      code.addAload(1);
      //code.addCheckcast(pojoName);

      //First param
      code.addAload(2);
      code.addIconst(0);
      code.addOpcode(Opcode.AALOAD);
      code.addCheckcast("java/lang/String");
      
      //Second param
      code.addAload(2);
      code.addIconst(1);
      code.addOpcode(Opcode.AALOAD);
      code.addCheckcast("java/lang/Integer");
      code.addInvokevirtual("java/lang/Integer", "intValue", "()I");
      code.addInvokevirtual(pojoName, "accessInt", "(Ljava/lang/String;I)I");

      //return
      code.addInvokestatic("java/lang/Integer", "valueOf", "(I)Ljava/lang/Integer;");
      code.addOpcode(Opcode.ARETURN);

      code.setMaxLocals(3);

      CodeAttribute ca = code.toCodeAttribute();
      minfo.setCodeAttribute(ca);
      //StackMapTable.Writer writer = new StackMapTable.Writer(32);
      //writer.sameFrame(pc);
      //ca.setAttribute(writer.toStackMapTable(cp));
      cf.addMethod2(minfo);

      byte[] bytes = toBytecode(cf);
      return null;
//      return toClass(name, bytes, Thread.currentThread().getContextClassLoader(), null);
      //*** tested
   }
   
   
   private static byte[] toBytecode(ClassFile cf) throws IOException {
      ByteArrayOutputStream barray = new ByteArrayOutputStream();
      // javassist.bytecode.DataOutputStream out = new javassist.bytecode.DataOutputStream(barray);
      DataOutputStream out = new DataOutputStream(barray);
      try {
          cf.write(out);
      }
      finally {
          out.close();
      }

      return barray.toByteArray();
  }

   
   public static Class<?> toClass(String name, byte[] b, ClassLoader loader, ProtectionDomain domain)
         throws CannotCompileException
   {
      try
      {
         Method method;
         Object[] args;
         if (domain == null)
         {
            method = defineClass1;
            args = new Object[]
            {name, b, new Integer(0), new Integer(b.length)};
         }
         else
         {
            method = defineClass2;
            args = new Object[]
            {name, b, new Integer(0), new Integer(b.length), domain};
         }

         return toClass2(method, loader, args);
      }
      catch (RuntimeException e)
      {
         throw e;
      }
      catch (java.lang.reflect.InvocationTargetException e)
      {
         throw new CannotCompileException(e.getTargetException());
      }
      catch (Exception e)
      {
         throw new CannotCompileException(e);
      }
   }

   private static synchronized Class<?> toClass2(Method method, ClassLoader loader, Object[] args) throws Exception
   {
//      SecurityActions.setAccessible(method, true);
//      Class<?> clazz = (Class<?>) method.invoke(loader, args);
//      SecurityActions.setAccessible(method, false);
//      return clazz;
      
       method.setAccessible(true);
       Class<?> clazz = (Class<?>) method.invoke(loader, args);
       method.setAccessible(false);
       return clazz;
   }

   public Object invoke(Object target, Object[] params, long hash)
   {
      if (hash == 2)
         return ((POJO)target).accessInt((String)params[0], ((Integer)params[1]).intValue());
      if (hash == 1)
         return ((POJO)target).accessInt((String)params[0], ((Integer)params[1]).intValue());
      return null;
   }
}
